using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.Remoting;
using MacGruber;
using MeshVR;
using MVR.FileManagementSecure;
using Obi.Characters.ThirdPerson;
using SimpleJSON;
using UnityEngine;
using UnityEngine.UI;

namespace CheesyFX
{
    public class PoseMe : MVRScript
    {
        private const string baseStorePath = "Saves/PluginData/CheesyFX/BodyLanguage/PoseMe/";
        private static JSONStorableBool includeRoot = new JSONStorableBool("Include Root", true);
        private static JSONStorableBool includeBones = new JSONStorableBool("Include Bones", true);
        private static JSONStorableBool includeMorphs = new JSONStorableBool("Include Morphs", false);
        private static JSONStorableBool restoreHeadRotation = new JSONStorableBool("Restore Head Rotation", true);

        private static JSONStorableString baseName = new JSONStorableString("Base Name", "");
        
        private static JSONStorableBool autoStoreToDisk = new JSONStorableBool("Auto Store To Disk", true);
        private static JSONStorableBool createScreenShotOnUpdate = new JSONStorableBool("Create Screenshot On Update", false);

        private static List<JSONClass> poses = new List<JSONClass>();

        private JSONStorableStringChooser poseChooser =
            new JSONStorableStringChooser("poseChooser", new List<string>(), "", "Pose", ApplyPose);

        private Transform confirmCanvas;
        private static List<Atom> dildos = new List<Atom>();
        private string sceneType;
        private static IEnumerator applyPosePost;
        private static List<int> softPhysicsStates = new List<int>();
        private static List<Person> persons => FillMeUp.persons;
        private UnityEventsListener uiListener;


        public override void Init()
        {
            try
            {
                UIManager.SetScript(this, CreateUIElement, leftUIElements, rightUIElements);
                Utils.OnInitUI(CreateUIElement);
                FileManagerSecure.CreateDirectory(baseStorePath);
                this.SetupButton("Add Pose", false, AddPose);
                this.SetupButton("Update Pose", false, OverridePose);
                poseChooser.CreateUI(this, rightSide:true, chooserType:1);
                restoreHeadRotation.CreateUI(this);
                autoStoreToDisk.CreateUI(this);
                createScreenShotOnUpdate.CreateUI(this);

                Utils.SetupTextInput(this, baseName.name, baseName, true);

                baseName.val = sceneType = GetSceneContent();

                SuperController.singleton.onAtomAddedHandlers += OnAtomsChanged;
                SuperController.singleton.onAtomRemovedHandlers += OnAtomsChanged;
                // AddPose();
                // CreateAtomCo("Empty", "BLButtonHolder", val => empty = val).Start();
            }
            catch (Exception e)
            {
                SuperController.LogError(e.ToString());
            }
        }

        private void OnAtomsChanged(Atom atom)
        {
            var newSceneType = GetSceneContent();
            if(baseName.val == sceneType)
            {
                
                baseName.val = newSceneType;
            }

            sceneType = newSceneType;
        }
        
        public override void InitUI()
        {
            base.InitUI();
            if(UITransform == null || uiListener != null) return;
            uiListener = UITransform.gameObject.AddComponent<UnityEventsListener>();
            uiListener.onEnabled.AddListener(() => UIManager.SetScript(this, CreateUIElement, leftUIElements, rightUIElements));
            uiListener.onEnabled.AddListener(() => Utils.OnInitUI(CreateUIElement));
        }

        private string GetSceneContent()
        {
            int femCount = 0;
            int futaCount = 0;
            int maleCount = 0;
            dildos.Clear();
            foreach (var atom in SuperController.singleton.GetAtoms())
            {
                if (atom.type == "Person")
                {
                    var person = persons.First(x => x.uid == atom.uid);
                    if (person.genderListener.dcs.gender == DAZCharacterSelector.Gender.Female)
                    {
                        if (person.genderListener.isFuta) futaCount++;
                        else femCount++;
                    }
                    else maleCount++;
                }
                else if (atom.type == "Dildo")
                {
                    dildos.Add(atom);
                }
            }
            var sceneType = new string('F', femCount)+new string('M', maleCount)+new string('D', dildos.Count);
            return sceneType;
        }
        
        private JSONClass GetPose()
        {
            dildos.Clear();
            var pose = new JSONClass();
            pose["SceneType"] = sceneType;
            foreach (var atom in SuperController.singleton.GetAtoms())
            {
                if (atom.type == "Person")
                {
                    var person = persons.First(x => x.uid == atom.uid);
                    pose[atom.uid] = GetPersonPose(atom);
                    // if (person.genderListener.dcs.gender == DAZCharacterSelector.Gender.Female)
                    // {
                    //     if (person.genderListener.isFuta) futaCount++;
                    //     else femCount++;
                    // }
                    // else maleCount++;
                }
                else if (atom.type == "Dildo")
                {
                    // dildos.Add(atom);
                    pose[atom.uid] = GetDildoPose(atom);
                }
            }
            
            // sceneType = new string('F', femCount)+new string('M', maleCount)+new string('D', dildos.Count);
            
            return pose;
        }

        private string GetPosePath(int id)
        {
            return $"{baseStorePath}{sceneType}/{baseName.val}_{id:000}";
        }

        private void AddPose()
        {
            var pose = GetPose();
            poses.Add(pose);
            poseChooser.choices = Enumerable.Range(0, poses.Count).Select(x => x.ToString()).ToList();
            poseChooser.valNoCallback = poseChooser.choices.Last();
            FileManagerSecure.CreateDirectory(baseStorePath+pose["SceneType"]);
            if(autoStoreToDisk.val)
            {
                var path = GetPosePath(poseChooser.choices.Count - 1);
                // SavePose(pose, $"{baseStorePath}{pose["SceneType"].Value}/{pose["SceneType"].Value}_{poseChooser.choices.Count - 1:000}");
                StorePose(pose, path, autoStoreToDisk.val);
                $"Pose stored to {path}".Print();
            }
        }
        
        private void OverridePose()
        {
            var pose = GetPose();
            poses[int.Parse(poseChooser.val)] = pose;
            if(autoStoreToDisk.val)
            {
                var path = GetPosePath(int.Parse(poseChooser.val));
                StorePose(pose, path, createScreenShotOnUpdate.val);
                $"Pose {path} updated".Print();
            }
        }

        private static void ApplyPose(string poseId)
        {
            applyPosePost.Stop();
            softPhysicsStates.Clear();
            for (int i = 0; i < persons.Count; i++)
            {
                softPhysicsStates.Add(persons[i].GetSoftPhysicsState());
            }
            var pose = poses[int.Parse(poseId)];
            foreach (string atomUid in pose.Keys)
            {
                if(atomUid == "SceneType") continue;
                var atom = SuperController.singleton.GetAtomByUid(atomUid);
                if (atom == null)
                {
                    $"Atom {atomUid} not found".Print();
                    continue;
                }
                if(atom.type == "Person")
                {
                    Quaternion headCtrlRot = Quaternion.identity;
                    Quaternion headRot = Quaternion.identity;
                    Quaternion neckRot = Quaternion.identity;
                    Quaternion chestRot = Quaternion.identity;
                    FreeControllerV3 headCtrl = null;
                    DAZBone head = null;
                    DAZBone neck = null;
                    DAZBone chest = null;
                    if (!restoreHeadRotation.val)
                    {
                        headCtrl = atom.freeControllers.First(x => x.name == "headControl");
                        head = atom.GetComponentsInChildren<DAZBone>().FirstOrDefault(x => x.name == "head");
                        neck = atom.GetComponentsInChildren<DAZBone>().FirstOrDefault(x => x.name == "neck");
                        chest = atom.GetComponentsInChildren<DAZBone>().FirstOrDefault(x => x.name == "chest");
                        headCtrlRot = headCtrl.transform.localRotation;
                        headRot = head.transform.localRotation;
                        neckRot = neck.transform.localRotation;
                        chestRot = chest.transform.localRotation;
                    }
                    PresetManagerControl pmc = atom.presetManagerControls.First(x => x.name == "PosePresets");
                    var pm = pmc.GetComponent<PresetManager>();
                    pm.LoadPresetFromJSON(pose[atomUid].AsObject);
                    if (!restoreHeadRotation.val)
                    {
                        headCtrl.transform.localRotation = headCtrlRot;
                        head.transform.localRotation = headRot;
                        neck.transform.localRotation = neckRot;
                        chest.transform.localRotation = chestRot;
                    }
                }

                
                // else if (atom.type == "Dildo")
                // {
                //     if(loadDildoPoseCo == null) loadDildoPoseCo = LoadDildoPoseCo(pose.AsObject).Start();
                // }
            }
            applyPosePost = ApplyPosePost(pose).Start();
        }

        private static IEnumerator ApplyPosePost(JSONClass pose)
        {
            for (int i = 0; i < dildos.Count; i++)
            {
                dildos[i].collisionEnabled = false;
            }
            yield return new WaitUntil(() =>
            {
                for (int i = 0; i < persons.Count; i++)
                {
                    if (persons[i].GetSoftPhysicsState() != softPhysicsStates[i]) return false;
                }
                return true;
            });
            foreach (string atomUid in pose.Keys)
            {
                if(atomUid == "SceneType") continue;
                var atom = SuperController.singleton.GetAtomByUid(atomUid);
                var atomPose = pose[atomUid].AsObject;
                if (atom.type == "Person")
                {
                    JSONArray storables = atomPose["storables"].AsArray;
                    foreach (var bone in atom.GetComponentsInChildren<DAZBone>())
                    {
                        for (int i = 0; i < storables.Count; i++)
                        {
                            JSONClass jc = storables[i].AsObject;
                            if(jc["id"] != bone.name) continue;
                            bone.RestoreFromJSON(jc);
                        }
                    }
                }
                else if(atom.type == "Dildo")
                {
                    LoadDildoPose(atom, pose[atom.uid].AsObject);
                    atom.collisionEnabled = true;
                }
            }
        }

        private void StorePose(JSONClass jc, string path, bool doScreenShot = false)
        {
            SuperController.singleton.SaveJSON(jc, $"{path}.json");
            if(doScreenShot) SuperController.singleton.DoSaveScreenshot($"{path}.jpg");
        }
        
        private void StorePoses()
        {
            poses.ForEach(x => StorePose(x, GetPosePath(poses.IndexOf(x))));
        }

        
        private static bool ShouldStorableBeIncluded(JSONStorable t)
        {
            if (t is FreeControllerV3 && (includeRoot.val || ((FreeControllerV3) t).name != "control")) return true;//includeRoot.val || ((FreeControllerV3) t).name != "control";
            if (includeBones.val && t is DAZBone) return true;
            if (includeMorphs.val && t.storeId == "geometry" && t is DAZCharacterSelector) return true;
            return false;
        }

        // based on AcidBubbles Timeline
        private JSONClass GetPersonPose(Atom atom)
        {
            JSONClass poseJSON = null;
            var storables = atom.GetStorableIDs().Select(atom.GetStorableByID)
                .Where(t => !t.exclude && t.gameObject.activeInHierarchy)
                .Where(t => ShouldStorableBeIncluded(t));
            var storablesJSON = new JSONArray();
            foreach (var storable in storables)
            {
                var storableJSON = storable.GetJSON(true, false, storable.name != "geometry");
                storablesJSON.Add(storableJSON);
            }

            poseJSON = new JSONClass
            {
                // ["clipPoseSettings"] = new JSONClass
                // {
                //     ["includeRoot"] = { AsBool = includeRoot },
                //     ["includePose"] = { AsBool = includePose },
                //     ["includeMorphs"] = { AsBool = includeMorphs },
                //     ["useMergeLoad"] = { AsBool = useMergeLoad },
                // },
                ["setUnlistedParamsToDefault"] = { AsBool = true },
                ["V2"] = { AsBool = true },
                ["storables"] = storablesJSON
            };
            return poseJSON;
        }

        private JSONClass GetDildoPose(Atom atom)
        {
            JSONClass jc = new JSONClass();
            jc["control"] = atom.GetJSON(true, true, false);
            foreach (var rb in atom.GetComponentsInChildren<Rigidbody>())
            {
                var pos = rb.position.ToJA();
                var rot = rb.rotation.eulerAngles.ToJA();
                jc[rb.name] = new JSONClass
                {
                    ["pos"] = pos,
                    ["rot"] = rot
                };
            }
            return jc;
        }

        private static void LoadDildoPose(Atom atom, JSONClass pose)
        {
            atom.RestoreFromJSON(pose["control"].AsObject);
            foreach (var rb in atom.GetComponentsInChildren<Rigidbody>())
            {
                rb.position = pose[rb.name]["pos"].AsArray.ToV3();
                rb.rotation = Quaternion.Euler(pose[rb.name]["rot"].AsArray.ToV3());
            }
        }
        
        public static IEnumerator CreateAtomCo(string type, string basename, Action<Atom> onAtomCreated)
        {
            string uid = NewUID(basename);
            yield return SuperController.singleton.AddAtomByType(type, uid, true);
            var atom = SuperController.singleton.GetAtomByUid(uid);
            if(atom == null)
            {
                throw new NullReferenceException("atom was not created");
            }
            SuperController.singleton.RenameAtom(atom, basename);
            onAtomCreated(atom);
        }

        public static string NewUID(string basename)
        {
            var uids = new HashSet<string>(SuperController.singleton.GetAtomUIDs());
            if(!uids.Contains(basename))
            {
                return basename;
            }

            for(int i = 2; i < 1000; i++)
            {
                string uid = $"{basename}#{i}";
                if(!uids.Contains(uid))
                {
                    return uid;
                }
            }

            // you don't really want to be here 
            return basename + Guid.NewGuid();
        }
        
        public void ForceSaveJSON(JSONClass jc, string path)
        {
            
            var confirmButtonTransform = SuperController.singleton.transform
                .Find("WorldScaleAdjust/HUDForLogs/UserConfirmCanvas/UserConfirmPluginActionPanel(Clone)/Panel/ConfirmButton");
            if(confirmButtonTransform == null) return;
            55.Print();
            confirmButtonTransform.GetComponent<Button>().onClick.Invoke();
        }

        private void OnDestroy()
        {
            Destroy(uiListener);
            SuperController.singleton.onAtomAddedHandlers -= OnAtomsChanged;
            SuperController.singleton.onAtomRemovedHandlers -= OnAtomsChanged;
        }

        // StartCoroutine(CreateAtomCo("Empty", "shiny", atom => {
        //     // atom reference available here
        // }));
    }
}